/*
 *  g.cpp
 *  G
 *
 *  Created by Christopher Huff on Wed Oct 16 2002.
 *  Copyright (c) 2002 Christopher Huff. All rights reserved.
 *
 */

#include "g.h"
#include "gcompiler.h"
#include <cmath>

using namespace std;

namespace g {


double sqr(double n) {return n*n;}
double cub(double n) {return n*n*n;}
double min(double a, double b) {if(a < b) return a; else return b;}
double max(double a, double b) {if(a > b) return a; else return b;}

double vlengthsqr(double x, double y, double z) {return x*x + y*y + z*z;}
double vlength(double x, double y, double z) {return sqrt(vlengthsqr(x, y, z));}
void vnormalize(double * v, double x, double y, double z) {
    double tmp = vlengthsqr(x, y, z);
    v[0] /= tmp; v[1] /= tmp; v[2] /= tmp;
}
void vcross(GMachine & gmach)
{
    double * p1 = gmach.StackTopVector();
    double * p2 = gmach.StackPopVector();
    double tmp[3];
    tmp[0] = p1[0];
    tmp[1] = p1[1];
    tmp[2] = p1[2];
    p1[0] = tmp[1]*p2[2] - tmp[2]*p2[1];
    p1[1] = tmp[2]*p2[0] - tmp[0]*p2[2];
    p1[2] = tmp[0]*p2[1] - tmp[1]*p2[0];
}
void vdot(GMachine & gmach)
{
    double * p1 = gmach.StackTopVector();
    double * p2 = gmach.StackPopVector();
    gmach.StackPopScalar();
    gmach.StackPopScalar();
    gmach.StackTopScalar() = p1[0]*p2[0] + p1[1]*p2[1] + p1[2]*p2[2];
}
void vdist(GMachine & gmach)
{
    double * p1 = gmach.StackTopVector();
    double * p2 = gmach.StackPopVector();
    gmach.StackPopScalar();
    gmach.StackPopScalar();
    gmach.StackTopScalar() = vlength(p1[0] - p2[0], p1[1] - p2[1], p1[2] - p2[2]);
}

void SetupBuiltins(GMachine & gmach)
{
    #define MakeCFn(type, name) gmach.AddGlobal(#name, new type(gmach, #name, name));
    MakeCFn(FnS_S, cos)
    MakeCFn(FnS_S, sin)
    MakeCFn(FnS_S, tan)
    MakeCFn(FnS_S, acos)
    MakeCFn(FnS_S, asin)
    MakeCFn(FnS_S, atan)
    MakeCFn(FnS_SS, atan2)
    MakeCFn(FnS_S, cosh)
    MakeCFn(FnS_S, sinh)
    MakeCFn(FnS_S, tanh)
    
    MakeCFn(FnS_SS, pow)
    MakeCFn(FnS_S, sqrt)
    MakeCFn(FnS_S, sqr)
    MakeCFn(FnS_S, cub)
    MakeCFn(FnS_S, exp)
    MakeCFn(FnS_S, log)
    MakeCFn(FnS_S, log10)
    
    MakeCFn(FnS_S, ceil)
    MakeCFn(FnS_S, fabs)
    MakeCFn(FnS_S, floor)
    MakeCFn(FnS_SS, fmod)
    MakeCFn(FnS_SS, min)
    MakeCFn(FnS_SS, max)
    
    MakeCFn(FnS_V, vlengthsqr)
    MakeCFn(FnS_V, vlength)
    MakeCFn(FnV_V, vnormalize)
    MakeCFn(ScalarFunction, vcross)
    MakeCFn(ScalarFunction, vdot)
    MakeCFn(ScalarFunction, vdist)
    
    gmach.AddGlobal("version", G_VERSION);
    gmach.AddGlobal("pi", 3.14159265358979323846);
    gmach.AddGlobal("e", 2.7182818284590452354);
    gmach.AddGlobal("sqrt_2", sqrt(2));
    gmach.AddGlobal("x", 1, 0, 0);
    gmach.AddGlobal("y", 0, 1, 0);
    gmach.AddGlobal("z", 0, 0, 1);
    gmach.AddGlobal("o", 0, 0, 0);
}

GDataBlock::GDataBlock()
{
    start = new char[kBlockSize];
    endUsed = start;
    end = start + kBlockSize;
    size = kBlockSize;
}

GDataBlock::~GDataBlock() {
    delete[] start;
}

void GDataBlock::Resize()
{
    cout << "GDataBlock::Resize()" << endl;
    char * tmp = new char[size + kBlockSizeIncrement];
    char * c = start;
    char * dest = tmp;
    for(; c != endUsed; c++, dest++)
        *dest = *c;
    delete start;
    start = tmp;
}

size_t GDataBlock::AddData(double scal)
{
    size_t idx = endUsed - start;
    if(idx >= size) Resize();
    *((double *)endUsed) = scal;
    endUsed += sizeof(double);
    return idx;
}

size_t GDataBlock::AddData(double x, double y, double z)
{
    size_t idx = endUsed - start;
    if(idx >= size) Resize();
    ((double *)endUsed)[0] = x;
    ((double *)endUsed)[1] = y;
    ((double *)endUsed)[2] = z;
    endUsed += sizeof(double)*3;
    return idx;
}

size_t GDataBlock::AddData(GFunction * fn)
{
    size_t idx = endUsed - start;
    if(idx >= size) Resize();
    *((GFunction **)endUsed) = fn;
    endUsed += sizeof(GFunction *);
    return idx;
}


SymbolTable::SymbolTable()
{
    
}

SymbolTable::~SymbolTable()
{
    
}

const GSymbol * SymbolTable::Match(const std::string & name)
{
    GSymbol * res = NULL;
    
    int n = symbols.size();
    for(int j = 0; j < n; j++)
    {
        if(symbols[j].name == name)
        {
            res = &symbols[j];
            break;
        }
    }
    
    return res;
}


GMachine::GMachine()
{
    blocks[0] = new Block;
    numBlocks = 0;
    currentBlock = blocks[0];
    
    SetupBuiltins(*this);
}

GMachine::~GMachine()
{
    
}


void GMachine::Compile(const std::string & source)
{
    GCompiler compiler(*this);
    compiler.Compile(source);
}

void GMachine::Compile(std::istream & source)
{
    
}

GFunction * GMachine::GetFunction(const std::string & name)
{
    const GSymbol * sym = globalSymbols.Match(name);
    
    if(sym)
        return *((GFunction **)(globalData.start + sym->index));
    else
        return NULL;
}

size_t GMachine::AddGlobal(const std::string & name, double scal)
{
    size_t idx = globalData.AddData(scal);
    globalSymbols.AddSymbol(GSymbol(kScalarValue, name, kGlobalData, idx));
    return idx;
}

size_t GMachine::AddGlobal(const std::string & name, double x, double y, double z)
{
    size_t idx = globalData.AddData(x, y, z);
    globalSymbols.AddSymbol(GSymbol(kVectorValue, name, kGlobalData, idx));
    return idx;
}

size_t GMachine::AddGlobal(const std::string & name, GFunction * fn)
{
//    cout << "AddGlobal: " << name << endl;
    size_t idx = globalData.AddData(fn);
    switch(fn->Type()) {
        case kScalarValue:
            globalSymbols.AddSymbol(GSymbol(kScalarFunction, name, kGlobalData, idx));
        break;
        case kVectorValue:
            globalSymbols.AddSymbol(GSymbol(kVectorFunction, name, kGlobalData, idx));
        break;
        default:
            globalSymbols.AddSymbol(GSymbol(kNoReturnFunction, name, kGlobalData, idx));
    }
//    cout << "Global function added: " << name << ", index: " << idx << endl;
    return idx;
}


}//namespace g


