/*
 *  g.h
 *  G
 *
 *  Created by Christopher Huff on Wed Oct 16 2002.
 *  Copyright (c) 2002 Christopher Huff. All rights reserved.
 *
 */

#ifndef G_H
#define G_H

#include <string>
#include <iostream>
#include <vector>

namespace g {

#define G_VERSION_STRING "G version 0.1"
#define G_VERSION 0.1

class GFunction;
class GCompiler;

enum {
    kNoValue,
    kScalarValue,
    kVectorValue,
/*    kScalarConstant,
    kVectorConstant,*/
    kNoReturnFunction,
    kScalarFunction,
    kVectorFunction
};

enum {
    kLocalData,//Data local to the function
    kConstantData,//Constant data (local)
    kGlobalData//Global data
};

struct GSymbol {
    int type;
    int storage;
    int index;
    std::string name;
    GSymbol() {type = kNoValue; storage = kLocalData; index = 0;}
    GSymbol(int tp, const std::string & nm, int strg, size_t idx)
        {type = tp; name = nm; storage = strg; index = idx;}
};

class SymbolTable {
  private:
    std::vector<GSymbol> symbols;
    
  public:
    SymbolTable();
    ~SymbolTable();
    
    const GSymbol * Match(const std::string & name);
    void AddSymbol(const GSymbol & sym) {symbols.push_back(sym);}
    void Dump() {
        int n = symbols.size();
        for(int j = 0; j < n; j++)
            cout << symbols[j].name << ":" << symbols[j].storage << ":" << symbols[j].index << endl;
    }
};

const size_t kBlockSize = 32768;//Any one function call must fit in this
const int kMaxBlocks = 1024;

const size_t kBlockSizeIncrement = 32768;

struct GDataBlock {
    char * start;
    char * endUsed;
    char * end;
    size_t size;
    GDataBlock();
    ~GDataBlock();
    void Resize();
    size_t AddData(double scal);
    size_t AddData(double x, double y, double z);
    size_t AddData(GFunction * fn);
    size_t GetSize() {return size;}
};

struct GDataCounter {
    size_t size;
    GDataCounter() {size = 0;}
    void Resize();
    size_t AddData(double scal);
    size_t AddData(double x, double y, double z);
    size_t AddData(GFunction * fn);
    size_t RemoveData(double scal);
    size_t RemoveData(double x, double y, double z);
    size_t RemoveData(GFunction * fn);
    void Reset() {size = 0;}
    size_t GetSize() {return size;}
};

class GMachine {
  private:
    SymbolTable globalSymbols;
    
    //Some memory hacking here to squeeze out the last bit of performance, functions
    //inlined for speed.
    //Doing things this way can drastically reduce overhead of function calls:
    //To pass parameters to a function, the parameters are pushed onto the stack.
    //The stack data then becomes the beginning of the data array for the call,
    //with no copying necessary.
    //This is really going to need extensive testing, lots of room for error here.
    struct Block {
        char * start;
        char * endUsed;
        char * end;
        Block() {start = new char[kBlockSize]; endUsed = start; end = start + kBlockSize;}
        ~Block() {delete[] start;}
    };
    Block * blocks[kMaxBlocks];
    int numBlocks;
    Block * currentBlock;
    
    GDataBlock globalData;
    
    //This function is called to reserve memory for a G function when it is called. It is not
    //safe to use while the function is executing.
    //Parameters:
    //localSize: the parameter and local data memory space the function will use. (bytes)
    //stackSize: the max stack memory space the function can use. This space is not
    //marked as "used" until data is pushed onto it, this value is only used to ensure there
    //will be enough. (bytes)
    //paramSize: the portion of the local space used by parameters and thus already allocated. (bytes)
    //localStart: a reference to the local data pointer, set to the appropriate location.
    //stackStart: a reference to the function stack pointer, set to the appropriate location.
    //**returnSize: the size of the return value (bytes)
    inline void ReserveFrameMemory(int localSize, int stackSize, int paramSize, char * & localStart, char * & stackStart)
    {
        if((currentBlock->endUsed + ((localSize - paramSize) + stackSize)) > currentBlock->end)
        {
            cout << "DEBUG: Creating new VM memory block" << endl;
            //No room for function call on this block, make new block.
            //This is a relatively expensive operation, and should be done
            //as rarely as possible.
            numBlocks++;
            Block * newBlock = blocks[numBlocks] = new Block;
            
            //Copy params to new block
            char * paramSrc = currentBlock->endUsed - paramSize;
            for(int j = 0; j < paramSize; j++)
                newBlock->start[j] = paramSrc[j];
            
            //delete parameters that were copied to the new block.
            currentBlock->endUsed = paramSrc;//just reset end pointer
            
            currentBlock = newBlock;//switch to new block
            //Give function pointers to the local and stack data areas
            localStart = currentBlock->start;//parameters already copied
            stackStart = localStart + localSize;//stack data starts after local data
            currentBlock->endUsed = stackStart;
        }
        else
        {
            //Give function pointers to the local and stack data areas
            localStart = currentBlock->endUsed - paramSize;//local data overlaps the stack of calling function
            stackStart = localStart + localSize;//stack data starts after local data
            currentBlock->endUsed = stackStart;
        }
    }
    inline void FreeFrameMemory(int size) {
        if(currentBlock->endUsed == currentBlock->start)
        {
            if(numBlocks > 0)
            {
                    //Done with this block, free and go back to previous.
                    //A block is not freed when it goes empty, but when it goes past empty.
                    delete currentBlock;
                    numBlocks--;
                    currentBlock--;
                    FreeFrameMemory(size);
                    //TODO: Need to copy return values to previous block
            }
        }
        else
        {
            currentBlock->endUsed -= size;
        }
    }
    
  public:
    GMachine();
    ~GMachine();
    
    void Compile(const std::string & source);
    void Compile(std::istream & source);
    
    GFunction * GetFunction(const std::string & name);
    
    size_t AddGlobal(const std::string & name, double scal);
    size_t AddGlobal(const std::string & name, double x, double y, double z);
    size_t AddGlobal(const std::string & name, GFunction * fn);
    
    inline void StackPushFunction(GFunction * fn) {
        *(GFunction **)(currentBlock->endUsed) = fn;
        currentBlock->endUsed += sizeof(GFunction *);
    }
    inline void StackPushScalar(double scal) {
        *(double *)currentBlock->endUsed = scal;
        currentBlock->endUsed += sizeof(double);
    }
    inline void StackPushVector(double x, double y, double z) {
        ((double *)currentBlock->endUsed)[0] = x;
        ((double *)currentBlock->endUsed)[1] = y;
        ((double *)currentBlock->endUsed)[2] = z;
        currentBlock->endUsed += sizeof(double)*3;
    }
    
    //Maybe cheaper to keep pointer to top data on stack instead of subtracting from past-the-end?
    //Data popped from stack is safe until other data is pushed.
    inline GFunction * StackTopFunction() {return *(GFunction **)(currentBlock->endUsed - sizeof(GFunction *));}
    inline double & StackTopScalar() {return *(double *)(currentBlock->endUsed - sizeof(double));}
    inline double * StackTopVector() {return (double *)(currentBlock->endUsed - sizeof(double)*3);}
    
    inline GFunction * StackPopFunction() {return *(GFunction **)(currentBlock->endUsed -= sizeof(GFunction *));}
    inline double & StackPopScalar() {return *(double *)(currentBlock->endUsed -= sizeof(double));}
    inline double * StackPopVector() {return (double *)(currentBlock->endUsed -= sizeof(double)*3);}
    inline char * StackTop() {return currentBlock->endUsed;}
    
    void DumpGlobals() {globalSymbols.Dump();}
    
    friend class GCompiler;
    friend class GVMFunction;
};

}//namespace g

#endif
