/*
 *  gcompiler.cpp
 *  G
 *
 *  Created by Christopher Huff on Wed Oct 16 2002.
 *  Copyright (c) 2002 Christopher Huff. All rights reserved.
 *
 */

#include "gcompiler.h"
#include "gopcodes.h"
#include "gexpression.h"

#include <string>
#include <stdlib.h>

namespace g {
using namespace ExpressNodes;

GFunctionInfo::GFunctionInfo(GMachine & mach)
{
    function = new GVMFunction(mach);
    returnType = kNoValue;
    stackSize = 0;
    stackDepth = 0;
    parameterDataSize = 0;
    localDataSize = 0;
}


size_t GFunctionInfo::AddScalarConstant(double s)
{
    return function->code->constantData.AddData(s);
}
size_t GFunctionInfo::AddVectorConstant(double x, double y, double z)
{
    return function->code->constantData.AddData(x, y, z);
}
size_t GFunctionInfo::AddFunctionConstant(GFunction * fn)
{
    return function->code->constantData.AddData(fn);
}

size_t GFunctionInfo::ReserveScalarData()
{
    size_t tmp = localDataSize;
    localDataSize += sizeof(double);
    return tmp;
}

size_t GFunctionInfo::ReserveVectorData()
{
    size_t tmp = localDataSize;
    localDataSize += sizeof(double)*3;
    return tmp;
}
size_t GFunctionInfo::ReserveFunctionData()
{
    size_t tmp = localDataSize;
    localDataSize += sizeof(GFunction *);
    return tmp;
}



GCompiler::GCompiler(GMachine & mach)
    :machine(mach)
{
}

GCompiler::~GCompiler()
{
}

void GCompiler::Compile(const std::string & source)
{
//    cout << "Tokenizing\n";
    
    tokenizer = new GTokenizer(source);
    
//    cout << "Compiling\n";
    
//    cout << source << endl;
    while(!tokenizer->Done())
    {
        CompileDeclaration(kGlobalData);
        Require(kSemicolonToken);
    }
    
//    cout << "Done with pre-compile stage\n";
    //this stage will eventually execute an initialization function
    //to set up global variables
//    FinishCompile();
    
//    cout << "Done Compiling\n";
}


void GCompiler::Require(int tok)
{
    if(tokenizer->CurrentToken().id != tok)
    {
        cout << "Unexpected token \"" << tokenizer->CurrentToken().name << "\" found, \""
             << tokenizer->StringForToken(tok) << "\" expected." << endl;
        Error();
    }
    else
    {
        tokenizer->GetNextToken();
    }
}


bool GCompiler::Allow(int tok)
{
    if(tokenizer->CurrentToken().id == tok)
    {
        tokenizer->GetNextToken();
        return true;
    }
    
    return false;
}


void GCompiler::Error()
{
    exit(-1);
}


void GCompiler::CompileCommand()
{
    GFunctionInfo * fnInfo = functions.top();
    cout << ".";
    switch(tokenizer->CurrentToken().id)
    {
        case kScalarToken:
        case kVectorToken:
        case kFunctionToken:
            CompileDeclaration(kLocalData);
        break;
        
        case kIdentifierToken:
        {
            std::string name = tokenizer->CurrentToken().name;
            Require(kIdentifierToken);
            //Is assignment or function call
            if(tokenizer->CurrentToken().id == kLParenthesesToken)
            {
                //Is function call:
                //Identifer(
            }
            else if(tokenizer->CurrentToken().id == kEqualToken)
            {
                //Is assignment:
                //Identifer =
                Require(kEqualToken);
                CompileExpression(kScalarValue);
                const GSymbol * sym = ResolveSymbol(name);
                AddOp(new Set(sym->type, sym->storage, sym->index));
            }
            else
            {
                cout << "Error in CompileCommand, identifier not understood." << endl;
                Error();
            }
        }
        break;
        
        case kLCurlyBracketToken:
            Require(kLCurlyBracketToken);
            OpenScope();
            while(!tokenizer->Done() && tokenizer->CurrentToken().id != kRCurlyBracketToken)
            {
                CompileCommand();
                Require(kSemicolonToken);
            }
            CloseScope();
            Require(kRCurlyBracketToken);
        break;
        
        case kIfToken:
        {
            tokenizer->GetNextToken();
            Require(kLParenthesesToken);
            CompileExpression(kScalarValue);
            Require(kRParenthesesToken);
            
            CondJump * cond = new CondJump(false);
            AddOp(cond);
            
            CompileCommand();
            
            if(Allow(kElseToken))
            {
                Jump * jmpFalse = new Jump;
                AddOp(jmpFalse);
                AddOp(new JumpMarker(cond));
                CompileCommand();
                AddOp(new JumpMarker(jmpFalse));
            }
            else
            {
                AddOp(new JumpMarker(cond));
            }
        }
        break;
        
        case kWhileToken:
        {
            tokenizer->GetNextToken();
            
            CondJump * cond = new CondJump(false);
            Jump * loopBack = new Jump;
            
            AddOp(new JumpMarker(loopBack));
            
            Require(kLParenthesesToken);
            CompileExpression(kScalarValue);
            Require(kRParenthesesToken);
            
            AddOp(cond);
            
            CompileCommand();
            
            AddOp(loopBack);
            AddOp(new JumpMarker(cond));
        }
        break;
        
        case kDoToken:
        {
            tokenizer->GetNextToken();
            
            CondJump * cond = new CondJump(true);
            
            AddOp(new JumpMarker(cond));
            
            CompileCommand();
            
            Require(kWhileToken);
            Require(kLParenthesesToken);
            CompileExpression(kScalarValue);
            Require(kRParenthesesToken);
            
            AddOp(cond);
        }
        break;
        
/*        case kForToken:
        {
            tokenizer->GetNextToken();
            
            CondJump * cond = new CondJump(false);
            Jump * loopBack = new Jump;
            
            AddOp(new JumpMarker(loopBack));
            
            Require(kLParenthesesToken);
            CompileExpression(kScalarValue);
            Require(kRParenthesesToken);
            
            AddOp(cond);
            
            CompileCommand();
            
            AddOp(loopBack);
            AddOp(new JumpMarker(cond));
        }
        break;*/
        
        case kPrintToken:
            tokenizer->GetNextToken();
            AddOp(new Print(CompileExpression()));
        break;
        
        case kReturnToken:
            tokenizer->GetNextToken();
            
            if(fnInfo->returnType == kScalarValue)
                CompileExpression(kScalarValue);
            else if(fnInfo->returnType == kVectorValue)
                CompileExpression(kVectorValue);
            
            AddOp(new Return(fnInfo->returnType));
        break;
        
        default:
            cout << "Error in CompileCommand." << endl;
            Error();
    }
}


void GCompiler::CompileDeclaration(int storage)
{
//    cout << "Compiling declaration\n";
    if(tokenizer->CurrentToken().id == kFunctionToken)
    {
        CompileFunctionDeclaration(storage);
    }
    else if(tokenizer->CurrentToken().id == kScalarToken)
    {
        CompileNumericDeclaration(storage, kScalarValue);
    }
    else if(tokenizer->CurrentToken().id == kVectorToken)
    {
        CompileNumericDeclaration(storage, kVectorValue);
    }
}


void GCompiler::CompileFunctionDeclaration(int storage)
{
    int functionType = kNoReturnFunction;
    GFunctionInfo * fnInfo = new GFunctionInfo(machine);
    functions.push(fnInfo);
//    cout << "Compiling function definition\n";
    
    Require(kFunctionToken);
    
    if(tokenizer->CurrentToken().id == kScalarToken)
    {
        fnInfo->returnType = kScalarValue;
        functionType = kScalarFunction;
    }
    else if(tokenizer->CurrentToken().id == kVectorToken)
    {
        fnInfo->returnType = kVectorValue;
        functionType = kVectorFunction;
    }
    
    tokenizer->GetNextToken();
    
    fnInfo->name = tokenizer->CurrentToken().name;
    
    tokenizer->GetNextToken();
    
    OpenScope();//opens scope of parameters and deepest function-local variables
    //Parse the parameter list
    Require(kLParenthesesToken);
    
    while(tokenizer->CurrentToken().id != kRParenthesesToken)
    {
        int idx;
        string paramName;
        
        switch(tokenizer->CurrentToken().id)
        {
            case kScalarToken:
                tokenizer->GetNextToken();
                paramName = tokenizer->CurrentToken().name;
                idx = fnInfo->ReserveScalarData();
                AddSymbol(paramName, kScalarValue, kLocalData, idx);
                fnInfo->parameters.push_back(GSymbol(kScalarValue, paramName, kLocalData, idx));
                tokenizer->GetNextToken();
            break;
            
            case kVectorToken:
                tokenizer->GetNextToken();
                paramName = tokenizer->CurrentToken().name;
                idx = fnInfo->ReserveVectorData();
                AddSymbol(paramName, kVectorValue, kLocalData, idx);
                fnInfo->parameters.push_back(GSymbol(kScalarValue, paramName, kLocalData, idx));
                tokenizer->GetNextToken();
            break;
        }
        
        if(tokenizer->CurrentToken().id != kRParenthesesToken)
            Require(kCommaToken);
    }
    
    Require(kRParenthesesToken);//Get rid of the closing ")"
    
    //Add symbol (done before block, to allow recursion without a prototype)
    size_t idx;
    if(storage != kGlobalData)
        idx = functions.top()->AddFunctionConstant(fnInfo->function);
    else
        idx = machine.AddGlobal(fnInfo->name, fnInfo->function);
    
    AddSymbol(fnInfo->name, functionType, storage, idx);
    
    //Parse block
    if(Allow(kLCurlyBracketToken))
    {
        CompileFunctionCode();
        Require(kRCurlyBracketToken);
    }
    //else is prototype
    CloseScope();
    
    functions.pop();
    
//    fnInfo->function->Dump();
}

void GCompiler::CompileFunctionCode()
{
    GFunctionInfo * fnInfo = functions.top();
    //Compile source code into expression list/tree structure
    while(!tokenizer->Done() && tokenizer->CurrentToken().id != kRCurlyBracketToken)
    {
        CompileCommand();
        Require(kSemicolonToken);
    }
    
    //Optimize expression trees
    
    //Process expression trees into opcode list
//    cout << "Generating opcodes\n";
    int n = fnInfo->expressionTrees.size();
    for(int j = 0; j < n; j++)
        fnInfo->expressionTrees[j]->EmitCode(*fnInfo);
    
    //Copy over code and data into final function code structure
//    cout << "Creating function\n";
    GFunctionCode * fnCode = fnInfo->function->code;
    
    fnInfo->function->name = fnInfo->name;
    fnInfo->function->returnType = fnInfo->returnType;
    
    n = fnInfo->opcodes.size();
    fnCode->codeSize = n;
    opcode * ops = new opcode[n];
    fnCode->codes = ops;
    for(int j = 0; j < n; j++)
        ops[j] = fnInfo->opcodes[j];
    
    n = fnInfo->parameters.size();
    fnCode->numParameters = n;
    fnCode->parameters = new GSymbol[n];
    for(int j = 0; j < n; j++)
    {
        fnCode->parameters[j] = fnInfo->parameters[j];
        switch(fnInfo->parameters[j].type)
        {
            case kScalarValue: fnInfo->parameterDataSize += sizeof(double); break;
            case kVectorValue: fnInfo->parameterDataSize += sizeof(double)*3; break;
            case kNoReturnFunction:
            case kScalarFunction:
            case kVectorFunction:
                fnInfo->parameterDataSize += sizeof(GFunction *);
            break;
        
            default:
                cout << "Error in CompileFunctionCode, unknown parameter type." << endl;
                Error();
        }
    }
    
    fnCode->stackSize = fnInfo->stackSize;
    fnCode->parameterDataSize = fnInfo->parameterDataSize;
    fnCode->localDataSize = fnInfo->localDataSize;
    
/*    cout << "Stack Size: " << fnInfo->stackSize << endl;
    cout << "Stack Depth: " << fnInfo->stackDepth << endl;
    cout << "Parameter Size: " << fnInfo->parameterDataSize << endl;
    cout << "Data Size: " << fnInfo->localDataSize << endl;*/
}

//Compile vector or scalar declarations
void GCompiler::CompileNumericDeclaration(int storage, int type)
{
    GFunctionInfo * fnInfo = functions.top();
    tokenizer->GetNextToken();
    std::string name = tokenizer->CurrentToken().name;
    tokenizer->GetNextToken();
    
    if(storage == kLocalData)
    {
        size_t idx;
        if(type == kScalarValue)
            idx = fnInfo->ReserveScalarData();
        else
            idx = fnInfo->ReserveVectorData();
        
        AddSymbol(name, type, storage, idx);
        
        if(Allow(kEqualToken))
        {
            CompileExpression(type);
            AddOp(new Set(type, storage, idx));
        }
    }
    else
    {
        cout << "Global numeric variables not implemented yet." << endl;
        Error();
/*        if(Allow(kEqualToken))
        {
            cout << "Error: only local variables may be initialized at declaration." << endl;
            Error();
        }
        else
        {
            int idx = machine.globalScalarData.size();
            machine.globalScalarData.push_back(0);
            AddVariable(kScalarValue, fnInfo->name, idx);
        }*/
    }
}

//Compiles an expression of a desired type, converts result if necessary
//Possible optimization: convert values before computation, doing scalar math
//where vector math is unnecessary.
int GCompiler::CompileExpression(int wantedType)
{
    GFunctionInfo * fnInfo = functions.top();
    ArithmeticNode * expRoot = CompileExpressionTree();
    
    fnInfo->expressionTrees.push_back(expRoot);
    
    int type = expRoot->Type();
    
    if(wantedType != kNoValue)
    {
        if(type == kVectorValue && wantedType == kScalarValue)
            fnInfo->expressionTrees.push_back(new Make(kScalarValue));
        else if(type == kScalarValue && wantedType == kVectorValue)
            fnInfo->expressionTrees.push_back(new Make(kVectorValue));
    }
    
//    expRoot->Print();
//    cout << endl;
    
    return type;
}

//Compiles a numeric expression, returns type
ArithmeticNode * GCompiler::CompileExpressionTree()
{
    ArithmeticNode * root = new ArithmeticNode;
    ArithmeticNode * current = CompileSingleValue();
    
    root->SetRHS(current);
    
    while(IsOperation())
    {
        ArithmeticNode * opNode;
        switch(tokenizer->CurrentToken().id)
        {
            case kPlusToken: opNode = new AddNode; break;
            case kMinusToken: opNode = new SubNode; break;
            case kStarToken: opNode = new MultNode; break;
            case kSlashToken: opNode = new DivNode; break;
            case kLAngleBracketToken: opNode = new LessNode; break;
            case kRAngleBracketToken: opNode = new GreaterNode; break;
            default:
                cout << "Error in CompileExpression, unrecognized operator: "
                     << tokenizer->CurrentToken().name << endl;
                Error();
        }
        tokenizer->GetNextToken();
        
        opNode->SetRHS(CompileSingleValue());
        
        while(current->GetParent()->Precedence() != -1 && current->GetParent()->Precedence() <= opNode->Precedence())
            current = current->GetParent();
        
        current->GetParent()->SetRHS(opNode);
        opNode->SetLHS(current);
        current = opNode->GetRHS();
    }
    return root;
}

bool GCompiler::IsOperation()
{
    int id = tokenizer->CurrentToken().id;
    return id == kPlusToken ||
           id == kMinusToken ||
           id == kStarToken ||
           id == kSlashToken ||
           id == kLAngleBracketToken ||
           id == kRAngleBracketToken;
}

//Compiles a single value: a variable, constant, function call, etc
ArithmeticNode * GCompiler::CompileSingleValue()
{
    GFunctionInfo * fnInfo = functions.top();
    ArithmeticNode * val = NULL;
    int type = kNoValue;
    
    switch(tokenizer->CurrentToken().id)
    {
        case kIdentifierToken:
        {
            //Variable access or function call
            const GSymbol * sym = ResolveSymbol(tokenizer->CurrentToken().name);
            tokenizer->GetNextToken();
            if(sym->type == kNoReturnFunction || sym->type == kScalarFunction || sym->type == kVectorFunction)
            {
                if(sym->type == kScalarFunction)
                    type = kScalarValue;
                else if(sym->type == kVectorFunction)
                    type = kVectorValue;
                
                CallFunctionNode * fnCall = new CallFunctionNode(type, sym->storage, sym->index);
                val = fnCall;
                
                Require(kLParenthesesToken);
                if(!Allow(kRParenthesesToken))
                {
                    do
                        fnCall->AddParameter(CompileExpressionTree());
                    while(Allow(kCommaToken));
                    
                    Require(kRParenthesesToken);
                }
            }
            else
            {
                val = new AccessDataNode(sym->type, sym->storage, sym->index);
                type = sym->type;
            }
        }
        break;
        case kScalarValueToken:
        {
            double v = strtod(tokenizer->CurrentToken().name.c_str(), NULL);
            tokenizer->GetNextToken();
            val = new AccessDataNode(kScalarValue, kConstantData, fnInfo->AddScalarConstant(v));
            type = kScalarValue;
        }
        break;
        case kLParenthesesToken:
        {
            Require(kLParenthesesToken);
            val = CompileExpressionTree();
            Require(kRParenthesesToken);
            type = val->Type();
        }
        break;
/*        case kLAngleBracketToken:
        {
            Require(kLAngleBracketToken);*/
        case kLSquareBracketToken:
        {
            Require(kLSquareBracketToken);
            ArithmeticNode * xVal = CompileExpressionTree();
            Require(kCommaToken);
            ArithmeticNode * yVal = CompileExpressionTree();
            Require(kCommaToken);
            ArithmeticNode * zVal = CompileExpressionTree();
            Require(kRSquareBracketToken);
//            Require(kRAngleBracketToken);
            val = new VectorValueNode(xVal, yVal, zVal);
            type = kVectorValue;
        }
        break;
        default:
            cout << "Expected numeric value, \"" << tokenizer->CurrentToken().name
                 << "\" found instead." << endl;
            Error();
    }
    
    //Should make this an operator and move into expression parsing code
    if(tokenizer->CurrentToken().id == kDotToken)
    {
        Require(kDotToken);
        if(type == kVectorValue)
        {
            if(tokenizer->CurrentToken().name == "x" ||
               tokenizer->CurrentToken().name == "y" ||
               tokenizer->CurrentToken().name == "z")
            {
                val = new VectorComponentNode(val, tokenizer->CurrentToken().name[0]);
                tokenizer->GetNextToken();
            }
            else
            {
                cout << "Dot component not recognized. x, y, and z are valid, \""
                     << tokenizer->CurrentToken().name << "\" found." << endl;
                Error();
            }
        }
        else
        {
            cout << "Can not take vector component of non vector." << endl;
            Error();
        }
    }
    
    return val;
}

//Compiles an identifier symbol. Resolves, pushes onto appropriate stack, returns type.
int GCompiler::CompileSymbol()
{
    const GSymbol * sym = ResolveSymbol(tokenizer->CurrentToken().name);
    functions.top()->expressionTrees.push_back(new AccessDataNode(sym->type, sym->storage, sym->index));
    tokenizer->GetNextToken();
    return sym->type;
}

void GCompiler::OpenScope()
{
    SymbolTable tmp;
    functions.top()->symbolTables.push_front(tmp);
}

void GCompiler::CloseScope()
{
//    functions.top()->symbolTables.front().Dump();
    functions.top()->symbolTables.pop_front();
}

void GCompiler::AddSymbol(const string & name, int type, int storage, size_t idx)
{
    if(storage == kGlobalData)
    {
        machine.globalSymbols.AddSymbol(GSymbol(type, name, storage, idx));
    }
    else if(storage == kLocalData || storage == kConstantData)
    {
        functions.top()->symbolTables.front().AddSymbol(GSymbol(type, name, storage, idx));
    }
}

const GSymbol * GCompiler::ResolveSymbol(const std::string & name)
{
    const GSymbol * sym = NULL;
    GFunctionInfo * fnInfo = functions.top();
    
    //try local symbols first (local variables and constants)
    std::list<SymbolTable>::iterator i;
    for(i = fnInfo->symbolTables.begin(); i != fnInfo->symbolTables.end(); i++)
        if(sym = i->Match(name))
            break;
    
    if(!sym)
        sym = machine.globalSymbols.Match(name);
    
    if(!sym)
    {
        cout << "Error in GCompiler::ResolveSymbol(): Could not find symbol \""
             << name << "\", make sure it exists and is in scope." << endl;
        Error();
    }
/*    if(sym)
        cout << "Symbol \"" << name << "\" resolved, type: " << sym->type << ", data index: " << sym->index << endl;*/
    
    return sym;
}

void GCompiler::AddOp(ExpressNodes::ExpNode * op)
{
    functions.top()->expressionTrees.push_back(op);
}

}//namespace g

