/*
 *  gexpression.h
 *  G
 *
 *  Created by Christopher Huff on Mon Oct 21 2002.
 *  Copyright (c) 2002 Christopher Huff. All rights reserved.
 *
 */


#ifndef GEXPRESSION_H
#define GEXPRESSION_H

#include "gopcodes.h"
#include "gfunction.h"
#include "gcompiler.h"

namespace g {

namespace ExpressNodes {

class ExpNode {
  private:
  public:
    ExpNode() {}
    virtual ~ExpNode() {}
    
    virtual void EmitCode(GFunctionInfo & fninfo) {fninfo.AddOp(kNoOp);}
};

//Nodes that "have a value", the code they emit leaves a vector or
//scalar value on the stack.
class ArithmeticNode: public ExpNode {
  protected:
    ArithmeticNode * lhs;
    ArithmeticNode * rhs;
    ArithmeticNode * parent;
  public:
    ArithmeticNode() {lhs = NULL; rhs = NULL; parent = NULL;}
    ~ArithmeticNode() {if(lhs) delete lhs; if(rhs) delete rhs;}
    virtual int Precedence() {return -1;}//Precedence is reverse of numeric value: 1 is above 2
//    virtual bool IsConstant();//will eventually be used for optimization
    virtual int Type() {//FIXME
        if((lhs && lhs->Type() == kVectorValue) || (rhs && rhs->Type() == kVectorValue))
            return kVectorValue;
        else
            return kScalarValue;
    }
    void SetLHS(ArithmeticNode * nd) {lhs = nd; if(lhs) lhs->SetParent(this);}
    ArithmeticNode * GetLHS() {return lhs;}
    void SetRHS(ArithmeticNode * nd) {rhs = nd; if(rhs) rhs->SetParent(this);}
    ArithmeticNode * GetRHS() {return rhs;}
    void SetParent(ArithmeticNode * nd) {parent = nd;}
    ArithmeticNode * GetParent() {return parent;}
    
    void EmitCode(GFunctionInfo & fninfo) {rhs->EmitCode(fninfo);}
    
    virtual void Print() {cout << "{"; rhs->Print(); cout << "}";}
};

class VectorValueNode: public ArithmeticNode {
  protected:
    ArithmeticNode * xExpress;
    ArithmeticNode * yExpress;
    ArithmeticNode * zExpress;
  public:
    VectorValueNode(ArithmeticNode * xExp, ArithmeticNode * yExp, ArithmeticNode * zExp) {
        xExpress = xExp;
        yExpress = yExp;
        zExpress = zExp;
    }
    
    ~VectorValueNode() {
        if(xExpress) delete xExpress;
        if(yExpress) delete yExpress;
        if(zExpress) delete zExpress;
    }
    
    int Type() {return kVectorValue;}
    
    void EmitCode(GFunctionInfo & fninfo) {
        xExpress->EmitCode(fninfo);
        yExpress->EmitCode(fninfo);
        zExpress->EmitCode(fninfo);
    }
    
    virtual void Print() {
        cout << "<";
        xExpress->Print();
        cout << ", ";
        yExpress->Print();
        cout << ", ";
        zExpress->Print();
        cout << ">";
    }
};

//Takes information from a specific location in local, global, or constant data and
//pushes it onto the stack.
class AccessDataNode: public ArithmeticNode {
  protected:
    int type;
    size_t index;
    int storage;
  public:
    AccessDataNode(int tp, int strg, size_t idx) {type = tp; index = idx; storage = strg;}
    
    virtual int Precedence() {return 0;}
    
    virtual int Type() {return type;}
    
    void EmitCode(GFunctionInfo & fninfo) {
        switch(storage) {
            case kGlobalData:
                if(type == kScalarValue) fninfo.AddOp(kPushScalarGlobalOp | (index << 10));
                else if(type == kVectorValue) fninfo.AddOp(kPushVectorGlobalOp | (index << 10));
            break;
            case kLocalData:
                if(type == kScalarValue) fninfo.AddOp(kPushScalarLocalOp | (index << 10));
                else if(type == kVectorValue) fninfo.AddOp(kPushVectorLocalOp | (index << 10));
            break;
            case kConstantData:
                if(type == kScalarValue) fninfo.AddOp(kPushScalarConstantOp | (index << 10));
                else if(type == kVectorValue) fninfo.AddOp(kPushVectorConstantOp | (index << 10));
            break;
        }
        if(type == kScalarValue)
            fninfo.StackPushScalar();
        else if(type == kVectorValue)
            fninfo.StackPushVector();
        else
            {cout << "(AccessDataNode) Error in generating opcodes: type not recognized" << endl; exit(-1);}
    }
    
    virtual void Print() {
        switch(storage) {
            case kGlobalData: cout << "g"; break;
            case kLocalData: cout << "l"; break;
            case kConstantData: cout << "c"; break;
        }
        
        if(type == kScalarValue) cout << "s";
        else if(type == kVectorValue) cout << "v";
    }
};


//This node grabs a component of a vector
class VectorComponentNode: public ArithmeticNode {
  private:
    char component;
    ArithmeticNode * value;
  public:
    VectorComponentNode(ArithmeticNode * val, char comp) {component = comp; value = val;}
    
    virtual int Precedence() {return 0;}
    
    virtual int Type() {return kScalarValue;}
    
    void EmitCode(GFunctionInfo & fninfo) {
        value->EmitCode(fninfo);
        switch(component) {
            case 'x': fninfo.AddOp(kGetXOp); break;
            case 'y': fninfo.AddOp(kGetYOp); break;
            case 'z': fninfo.AddOp(kGetZOp); break;
        }
    }
    
    virtual void Print() {
        value->Print();
        cout << "." << component;
    }
};


class CallFunctionNode: public ArithmeticNode {
  protected:
    int type;
    size_t index;
    int storage;
    std::vector<ArithmeticNode *> params;
  public:
    CallFunctionNode(int tp, int strg, size_t idx) {type = tp; index = idx; storage = strg;}
    
    void AddParameter(ArithmeticNode * param) {params.push_back(param);}
    
    virtual int Precedence() {return 0;}
    
    virtual int Type() {return type;}
    
    void EmitCode(GFunctionInfo & fninfo) {
//        cout << "Function Index: " << index << endl;
        int n = params.size();
        for(int j = 0; j < n; j++)
            params[j]->EmitCode(fninfo);
        
        switch(storage) {
            case kGlobalData:
                fninfo.AddOp(kPushFunctionGlobalOp | (index << 10));
            break;
            case kLocalData:
                fninfo.AddOp(kPushFunctionLocalOp | (index << 10));
            break;
            case kConstantData:
                fninfo.AddOp(kPushFunctionConstantOp | (index << 10));
            break;
            default:
                {cout << "(CallFunctionNode) Error in generating opcodes: storage mode not recognized" << endl; exit(-1);}
        }
        fninfo.StackPushFunction();
        fninfo.AddOp(kFunctionCallOp);
        fninfo.StackPopFunction();
        if(type == kScalarValue)
            fninfo.StackPushScalar();
        else if(type == kVectorValue)
            fninfo.StackPushVector();
    }
    
    virtual void Print() {
        switch(storage) {
            case kGlobalData: cout << "g"; break;
            case kLocalData: cout << "l"; break;
            case kConstantData: cout << "c"; break;
        }
        
        if(type == kScalarValue) cout << "s";
        else if(type == kVectorValue) cout << "v";
        
        cout << "f(";
        int n = params.size();
        for(int j = 0; j < n; j++)
        {
            params[j]->Print();
            if(j != n - 1)
                cout << ", ";
        }
        cout << ")";
    }
};


#define BinaryOpNode(op, str, prec) \
class op##Node: public ArithmeticNode {\
  public:\
    virtual int Precedence() {return prec;}\
    void EmitCode(GFunctionInfo & fninfo) {\
        lhs->EmitCode(fninfo); rhs->EmitCode(fninfo);\
        if(lhs->Type() == kScalarValue && rhs->Type() == kScalarValue)\
            fninfo.AddOp(k##op##SSOp);\
        if(lhs->Type() == kScalarValue && rhs->Type() == kVectorValue)\
            fninfo.AddOp(k##op##SVOp);\
        if(lhs->Type() == kVectorValue && rhs->Type() == kScalarValue)\
            fninfo.AddOp(k##op##VSOp);\
        if(lhs->Type() == kVectorValue && rhs->Type() == kVectorValue)\
            fninfo.AddOp(k##op##VVOp);\
    }\
    virtual void Print() {cout << "("; lhs->Print(); cout << #str; rhs->Print(); cout << ")";}\
};

BinaryOpNode(Less, <, 3)
BinaryOpNode(Greater, >, 3)
BinaryOpNode(Add, +, 2)
BinaryOpNode(Sub, -, 2)
BinaryOpNode(Mult, *, 1)
BinaryOpNode(Div, /, 1)


class Make: public ExpNode {
  private:
    int type;
  public:
    Make(int datType) {type = datType;}
    void EmitCode(GFunctionInfo & fninfo) {
        if(type == kScalarValue) {
            fninfo.AddOp(kMakeScalarOp);
            fninfo.StackPopScalar();
            fninfo.StackPopScalar();
        }
        else if(type == kVectorValue) {
            fninfo.AddOp(kMakeVectorOp);
            fninfo.StackPushScalar();
            fninfo.StackPushScalar();
        }
        else {cout << "(Make) Error in generating opcodes: type not recognized" << endl; exit(-1);}
    }
};

//Sets a numeric value of type "type" at location "index"
//to the value of the same type on the stack
class Set: public ExpNode {
  private:
    size_t index;
    int storage;
    int type;
  public:
    Set(int datType, int strg, size_t datIdx) {type = datType; storage = strg; index = datIdx;}
    void EmitCode(GFunctionInfo & fninfo) {
        if(type == kScalarValue) {
            if(storage == kLocalData)
                fninfo.AddOp(kSetLocalScalarOp | (index << 10));
            else
                fninfo.AddOp(kSetGlobalScalarOp | (index << 10));
        }
        else if(type == kVectorValue) {
            if(storage == kLocalData)
                fninfo.AddOp(kSetLocalVectorOp | (index << 10));
            else
                fninfo.AddOp(kSetGlobalVectorOp | (index << 10));
        }
        else {cout << "(Set) Error in generating opcodes: type not recognized" << endl; exit(-1);}
        fninfo.StackPopVector();
    }
};


//Prints the top data on the stack to standard output
class Print: public ExpNode {
  private:
    int type;
  public:
    Print(int datType) {type = datType;}
    void EmitCode(GFunctionInfo & fninfo) {
        if(type == kScalarValue) {
            fninfo.AddOp(kPrintScalarOp);
            fninfo.StackPopScalar();
        }
        else if(type == kVectorValue) {
            fninfo.AddOp(kPrintVectorOp);
            fninfo.StackPopVector();
        }
        else {cout << "(Print) Error in generating opcodes: type not recognized" << endl; exit(-1);}
    }
};


class StackPop: public ExpNode {
  private:
    int type;
  public:
    StackPop(int datType) {type = datType;}
    void EmitCode(GFunctionInfo & fninfo) {
        if(type == kScalarValue) {
            fninfo.AddOp(kPopScalarOp);
            fninfo.StackPopScalar();
        }
        else if(type == kVectorValue) {
            fninfo.AddOp(kPopVectorOp);
            fninfo.StackPopVector();
        }
        else {cout << "(StackPop) Error in generating opcodes: type not recognized" << endl; exit(-1);}
    }
};


//index - 1 to jump to the point just before start point, pc will be incremented
class Jump: public ExpNode {
  protected:
    int index;//jump to location
    int codePosition;//location of jump op
  public:
    Jump(int codeIdx = 0) {index = codeIdx; codePosition = -1;}
    void SetIndex(int codeIdx) {index = codeIdx;}
    virtual void ResetIndex(GFunctionInfo & fninfo, int codeIdx) {
        index = codeIdx;
        if(codePosition != -1)//Don't try to modify code if not emitted yet
            fninfo.SetOp(codePosition, kJumpOp | ((index - 1) << 10));
    }
    void EmitCode(GFunctionInfo & fninfo) {
        codePosition = fninfo.Position();
        fninfo.AddOp(kJumpOp | ((index - 1) << 10));
    }
};


class CondJump: public Jump {
  protected:
    bool jumpCondition;
  public:
    CondJump(bool cond, int codeIdx = 0): Jump(codeIdx) {jumpCondition = cond;}
    virtual void ResetIndex(GFunctionInfo & fninfo, int codeIdx) {
        index = codeIdx;
        if(codePosition != -1)//Don't try to modify code if not emitted yet
            if(jumpCondition)
                fninfo.SetOp(codePosition, kCondJumpOp | ((index - 1) << 10));
            else
                fninfo.SetOp(codePosition, kInvCondJumpOp | ((index - 1) << 10));
    }
    void EmitCode(GFunctionInfo & fninfo) {
        codePosition = fninfo.Position();
        if(jumpCondition)
            fninfo.AddOp(kCondJumpOp | ((index - 1) << 10));
        else
            fninfo.AddOp(kInvCondJumpOp | ((index - 1) << 10));
    }
};


//This class doesn't emit any code, it just serves as a marker to set a
//jump operation to the correct location.
//Since it is usually encountered after the jump code has already been
//emitted, the jump op has to go back and modify what it emitted.
class JumpMarker: public ExpNode {
  protected:
    Jump * jumpOp;
  public:
    JumpMarker(Jump * jmp) {jumpOp = jmp;}
    void EmitCode(GFunctionInfo & fninfo) {jumpOp->ResetIndex(fninfo, fninfo.Position());}
};


class Return: public ExpNode {
  private:
    int type;
  public:
    Return(int retType) {type = retType;}
    void EmitCode(GFunctionInfo & fninfo) {
        if(type == kScalarValue) fninfo.AddOp(kSReturnOp);
        else if(type == kVectorValue) fninfo.AddOp(kVReturnOp);
        else fninfo.AddOp(kReturnOp);
    }
};

}//ExpressNodes

}//namespace g

#endif