/*
 *  gfunction.cpp
 *  G
 *
 *  Created by Christopher Huff on Wed Oct 16 2002.
 *  Copyright (c) 2002 Christopher Huff. All rights reserved.
 *
 */

#include "gfunction.h"
#include "gopcodes.h"

namespace g {

struct CodeStringPair {opcode code; const char * name;};

CodeStringPair codeStrings[kLastOp] = {
    {kAddSSOp, "+ss"},
    {kAddSVOp, "+sv"},
    {kAddVSOp, "+vs"},
    {kAddVVOp, "+vv"},
    {kSubSSOp, "-ss"},
    {kSubSVOp, "-sv"},
    {kSubVSOp, "-vs"},
    {kSubVVOp, "-vv"},
    {kMultSSOp, "*ss"},
    {kMultSVOp, "*sv"},
    {kMultVSOp, "*vs"},
    {kMultVVOp, "*vv"},
    {kDivSSOp, "/ss"},
    {kDivSVOp, "/sv"},
    {kDivVSOp, "/vs"},
    {kDivVVOp, "/vv"},
    {kLessSSOp, "<ss"},
    {kLessSVOp, "<sv"},
    {kLessVSOp, "<vs"},
    {kLessVVOp, "<vv"},
    {kGreaterSSOp, ">ss"},
    {kGreaterSVOp, ">sv"},
    {kGreaterVSOp, ">vs"},
    {kGreaterVVOp, ">vv"},
    {kMakeVectorOp, "mkvec"},
    {kMakeScalarOp, "mkscal"},
    {kSetLocalVectorOp, "setvl"},
    {kSetLocalScalarOp, "setsl"},
    {kSetGlobalVectorOp, "setvg"},
    {kSetGlobalScalarOp, "setsg"},
    {kPrintVectorOp, "printv"},
    {kPrintScalarOp, "prints"},
    {kPushFunctionLocalOp, "pushfl"},
    {kPushScalarLocalOp, "pushsl"},
    {kPushVectorLocalOp, "pushvl"},
    {kPushFunctionGlobalOp, "pushfg"},
    {kPushScalarGlobalOp, "pushsg"},
    {kPushVectorGlobalOp, "pushvg"},
    {kPushFunctionConstantOp, "pushfc"},
    {kPushScalarConstantOp, "pushsc"},
    {kPushVectorConstantOp, "pushvc"},
    {kPopScalarOp, "pops"},
    {kPopVectorOp, "popv"},
    {kJumpOp, "jmp"},
    {kCondJumpOp, "cjmp"},
    {kInvCondJumpOp, "icjmp"},
    {kFunctionCallOp, "fcall"},
    {kVFunctionCallOp, "vfcall"},
    {kSFunctionCallOp, "sfcall"},
    {kVReturnOp, "returnv"},
    {kSReturnOp, "returns"},
    {kReturnOp, "return"},
    {kStopOp, "stop"},
    {kNoOp, "noop"}
};

std::string OpcodeString(opcode cd)
{
    std::string codeStr;
    for(int j = 0; j < kLastOp; j++) {
        if(codeStrings[j].code == (cd & 1023)) {
            codeStr = codeStrings[j].name;
            break;
        }
    }
    
    return codeStr;
}


GFunctionCode::GFunctionCode() {
    codes = NULL;
    codeSize = 0;
    
    parameters = NULL;
    numParameters = 0;
    
    stackSize = 0;
    localDataSize = 0;
    parameterDataSize = 0;
}

GFunctionCode::~GFunctionCode() {
    delete[] codes;
    delete[] parameters;
}

void GVMFunction::Dump()
{
    cout << "Begin function code dump for function " << name << ":" << endl;
    
    for(int j = 0; j < code->numParameters; j++)
    {
        cout << code->parameters[j].name;
        if(j != code->numParameters - 1)
            cout << ", ";
    }
    
    cout << endl;
    cout << "numParameters = " << code->numParameters << endl;
    
    for(int j = 0; j < code->codeSize; j++)
        cout << OpcodeString(code->codes[j]) << ":" << (code->codes[j] >> 10) << endl;
}


GVMFunction::GVMFunction(GMachine & mach):
    GFunction(mach)
{
    code = new GFunctionCode;
}

GVMFunction::~GVMFunction()
{
    
}

void GVMFunction::Run()
{
    char * localData = NULL;//data array
    char * stackData = NULL;//function data array
    machine.ReserveFrameMemory(code->localDataSize, code->stackSize, code->parameterDataSize, localData, stackData);
    
/*    cout << "local data start:" << (int *)localData << endl;
    cout << "stack data start:" << (int *)stackData << endl;
    cout << "stack top:" << (int *)machine.StackTop() << endl;*/
    long op;//current op
    int opcode;//decoded code of current op
    long pc = 0;//Program counter
    bool go = true;//execution stops when this is false
    
    #define TopFunction machine.StackTopFunction()
    #define TopScalar machine.StackTopScalar()
    #define TopVector machine.StackTopVector()
    
    #define PopFunction machine.StackPopFunction()
    #define PopScalar machine.StackPopScalar()
    #define PopVector machine.StackPopVector()
    
    #define PushFunction(x) machine.StackPushFunction(x)
    #define PushScalar(x) machine.StackPushScalar(x)
    #define PushVector(x, y, z) machine.StackPushVector(x, y, z)
    
    
    #define LocalFunction(a) (*(GFunction **)(localData + (a)))
    #define LocalScalar(a) (*(double *)(localData + (a)))
    #define LocalVector(a) ((double *)(localData + (a)))
    
    #define ConstantFunction(a) (*(GFunction **)(code->constantData.start + (a)))
    #define ConstantScalar(a) (*(double *)(code->constantData.start + (a)))
    #define ConstantVector(a) ((double *)(code->constantData.start + (a)))
    
    #define GlobalFunction(a) (*(GFunction **)(machine.globalData.start + (a)))
    #define GlobalScalar(a) (*(double *)(machine.globalData.start + (a)))
    #define GlobalVector(a) ((double *)(machine.globalData.start + (a)))
    
    //these versions might be faster if the compiler is bad at optimizing,
    //but don't work for comparison operators.
/*    #define BinarySSOp(o) tmps0 = PopScalar; TopScalar o##= tmps0;
    
    #define BinaryVSOp(o) tmps0 = PopScalar; tmpv0 = TopVector; \
        tmpv0[0] o##= tmps0; \
        tmpv0[1] o##= tmps0; \
        tmpv0[2] o##= tmps0;
    
    //This macro does a little hacking to increase performance.
    //A vector is 3 scalars, so it can get away with only popping
    //one scalar off the stack (the Z value of the vector).
    #define BinarySVOp(o) tmps0 = PopScalar; tmpv0 = TopVector; tmps1 = tmpv0[0];\
        tmpv0[0] = tmps1 o tmpv0[1]; \
        tmpv0[1] = tmps1 o tmpv0[2]; \
        tmpv0[2] = tmps1 o tmps0;
    
    #define BinaryVVOp(o) tmpv1 = PopVector; tmpv0 = TopVector; \
        tmpv0[0] o##= tmpv1[0]; \
        tmpv0[1] o##= tmpv1[1]; \
        tmpv0[2] o##= tmpv1[2];
    */
    
    #define BinarySSOp(o) tmps0 = PopScalar; tmps1 = TopScalar; TopScalar = tmps1 o tmps0;
    
    #define BinaryVSOp(o) tmps0 = PopScalar; tmpv0 = TopVector; \
        tmpv0[0] = tmpv0[0] o tmps0; \
        tmpv0[1] = tmpv0[1] o tmps0; \
        tmpv0[2] = tmpv0[2] o tmps0;
    
    //This macro does a little hacking to increase performance.
    //A vector is 3 scalars, so it can get away with only popping
    //one scalar off the stack (the Z value of the vector).
    #define BinarySVOp(o) tmps0 = PopScalar; tmpv0 = TopVector; tmps1 = tmpv0[0];\
        tmpv0[0] = tmps1 o tmpv0[1]; \
        tmpv0[1] = tmps1 o tmpv0[2]; \
        tmpv0[2] = tmps1 o tmps0;
    
    #define BinaryVVOp(o) tmpv1 = PopVector; tmpv0 = TopVector; \
        tmpv0[0] = tmpv0[0] o tmpv1[0]; \
        tmpv0[1] = tmpv0[1] o tmpv1[1]; \
        tmpv0[2] = tmpv0[2] o tmpv1[2];
    
    //Some temporary variables
    double tmps0;
    double tmps1;
    double tmps2;
    double * tmpv0;
    double * tmpv1;
    
    while(go)
    {
        op = code->codes[pc];
        opcode = op & 1023;
        switch(opcode)
        {
            case kAddSSOp: BinarySSOp(+) break;
            case kAddSVOp: BinarySVOp(+) break;
            case kAddVSOp: BinaryVSOp(+) break;
            case kAddVVOp: BinaryVVOp(+) break;
            
            case kSubSSOp: BinarySSOp(-) break;
            case kSubSVOp: BinarySVOp(-) break;
            case kSubVSOp: BinaryVSOp(-) break;
            case kSubVVOp: BinaryVVOp(-) break;
            
            case kMultSSOp: BinarySSOp(*) break;
            case kMultSVOp: BinarySVOp(*) break;
            case kMultVSOp: BinaryVSOp(*) break;
            case kMultVVOp: BinaryVVOp(*) break;
            
            case kDivSSOp: BinarySSOp(/) break;
            case kDivSVOp: BinarySVOp(/) break;
            case kDivVSOp: BinaryVSOp(/) break;
            case kDivVVOp: BinaryVVOp(/) break;
            
            case kLessSSOp: BinarySSOp(<) break;
            case kLessSVOp: BinarySVOp(<) break;
            case kLessVSOp: BinaryVSOp(<) break;
            case kLessVVOp: BinaryVVOp(<) break;
            
            case kGreaterSSOp: BinarySSOp(>) break;
            case kGreaterSVOp: BinarySVOp(>) break;
            case kGreaterVSOp: BinaryVSOp(>) break;
            case kGreaterVVOp: BinaryVVOp(>) break;
            
            
            case kMakeScalarOp:
                //Convert a vector to a scalar (truncate, pop off the y and z)
                PopScalar; PopScalar;
            break;
            case kMakeVectorOp:
                //Convert a scalar to a vector with each component = the scalar
                //(push the scalar value twice)
                tmps0 = TopScalar;
                PushScalar(tmps0);
                PushScalar(tmps0);
            break;
            
            
            case kGetXOp:
                PopScalar; PopScalar;
            break;
            case kGetYOp:
                PopScalar; tmps0 = PopScalar; TopScalar = tmps0;
            break;
            case kGetZOp:
                tmps0 = PopScalar; PopScalar; TopScalar = tmps0;
            break;
            
            
            case kSetLocalScalarOp:
//                cout << "setsl " << (op >> 10) << ":" << ((double *)(localData + (op >> 10))) << " = " << TopScalar << endl;
                LocalScalar(op >> 10) = PopScalar;
            break;
            case kSetLocalVectorOp:
            {
                size_t address = op >> 10;
                tmpv0 = PopVector;
                LocalVector(address)[0] = tmpv0[0];
                LocalVector(address)[1] = tmpv0[1];
                LocalVector(address)[2] = tmpv0[2];
            }
            break;
            
            
            case kSetGlobalScalarOp:
                GlobalScalar(op >> 10) = TopScalar;
            break;
            case kSetGlobalVectorOp:
            {
                size_t address = op >> 10;
                tmpv0 = PopVector;
                GlobalVector(address)[0] = tmpv0[0];
                GlobalVector(address)[1] = tmpv0[1];
                GlobalVector(address)[2] = tmpv0[2];
            }
            break;
            
            
            case kPrintScalarOp:
                cout << PopScalar << endl;
            break;
            case kPrintVectorOp:
                tmpv0 = PopVector;
                cout << "< " << tmpv0[0] << ", " << tmpv0[1] << ", " << tmpv0[2] << ">" << endl;
            break;
            
            
            case kPushFunctionLocalOp:
                PushFunction(LocalFunction(op >> 10));
            break;
            case kPushScalarLocalOp:
//                cout << "pushsl " << (op >> 10) << ":" << ((double *)(localData + (op >> 10))) << " = " << LocalScalar(op >> 10) << endl;
                PushScalar(LocalScalar(op >> 10));
            break;
            case kPushVectorLocalOp:
            {
                size_t address = op >> 10;
                //could change to use PushVector?
                PushScalar(LocalVector(address)[0]);//x
                PushScalar(LocalVector(address)[1]);//y
                PushScalar(LocalVector(address)[2]);//z
            }
            break;
            
            
            case kPushFunctionConstantOp:
                PushFunction(ConstantFunction(op >> 10));
            break;
            case kPushScalarConstantOp:
//                cout << "pushsc " << (op >> 10) << " = " << ConstantScalar(op >> 10) << endl;
                PushScalar(ConstantScalar(op >> 10));
            break;
            case kPushVectorConstantOp:
            {
                size_t address = op >> 10;
                PushScalar(ConstantVector(address)[0]);//x
                PushScalar(ConstantVector(address)[1]);//y
                PushScalar(ConstantVector(address)[2]);//z
            }
            break;
            
            
            case kPushFunctionGlobalOp:
                PushFunction(GlobalFunction(op >> 10));
            break;
            case kPushScalarGlobalOp:
                PushScalar(GlobalScalar(op >> 10));
            break;
            case kPushVectorGlobalOp:
            {
                size_t address = op >> 10;
                PushScalar(GlobalVector(address)[0]);//x
                PushScalar(GlobalVector(address)[1]);//y
                PushScalar(GlobalVector(address)[2]);//z
            }
            break;
            
            
            case kPopScalarOp:
                PopScalar;
            break;
            case kPopVectorOp:
                PopVector;
            break;
            
            
            case kJumpOp:
                pc = op >> 10;
            break;
            case kCondJumpOp:
                if(PopScalar)
                    pc = op >> 10;
            break;
            case kInvCondJumpOp:
                if(!PopScalar)
                    pc = op >> 10;
            break;
            
            case kFunctionCallOp: {
                GFunction * fn = PopFunction;
                fn->Run();
            } break;
            
            //The data needs to be pulled off the stack and copied
            //back onto it after freeing the data.
            case kSReturnOp:
                tmps0 = PopScalar;
                machine.FreeFrameMemory(code->localDataSize);
                PushScalar(tmps0);
                go = false;
            break;
            case kVReturnOp:
                tmpv0 = PopVector;
                tmps0 = tmpv0[0];
                tmps1 = tmpv0[1];
                tmps2 = tmpv0[2];
                machine.FreeFrameMemory(code->localDataSize);
                PushVector(tmpv0[0], tmpv0[1], tmpv0[2]);
                go = false;
            break;
            case kReturnOp:
                machine.FreeFrameMemory(code->localDataSize);
                go = false;
            break;
            
            default:
                cout << "error: unrecognized opcode" << endl;
                exit(-1);
        }
        
        pc++;
    }
    
    return;
}



}//namespace g

