/*
 *  gfunction.h
 *  G
 *
 *  Created by Christopher Huff on Wed Oct 16 2002.
 *  Copyright (c) 2002 Christopher Huff. All rights reserved.
 *
 */

#ifndef GFUNCTION_H
#define GFUNCTION_H

#include <string>
#include <stack>
using std::stack;

#include "g.h"
//#include "gvector.h"

namespace g {

class GFunction {
  protected:
    GMachine & machine;
    std::string name;
    int returnType;
    
  public:
    GFunction(GMachine & mach): machine(mach) {}
    virtual ~GFunction() {}
    
    const std::string & Name() {return name;}
    int Type() {return returnType;}
    virtual void Run() = 0;
};

//Scalar return, scalar parameter
class FnS_S: public g::GFunction {
  protected:
    double (*fn)(double);
  public:
    FnS_S(g::GMachine & mach, const std::string & nm, double (*f)(double)): GFunction(mach)
    {name = nm; fn = f; returnType = g::kScalarValue;}
    virtual void Run() {machine.StackTopScalar() = fn(machine.StackTopScalar());}
};

//Scalar return, 2 scalar parameters
class FnS_SS: public g::GFunction {
  protected:
    double (*fn)(double, double);
  public:
    FnS_SS(g::GMachine & mach, const std::string & nm, double (*f)(double, double)): GFunction(mach)
    {name = nm; fn = f; returnType = g::kScalarValue;}
    virtual void Run() {
        double tmp = machine.StackPopScalar();
        machine.StackTopScalar() = fn(machine.StackTopScalar(), tmp);
    }
};

//Scalar return, vector parameter
class FnS_V: public g::GFunction {
  protected:
    double (*fn)(double, double, double);
  public:
    FnS_V(g::GMachine & mach, const std::string & nm, double (*f)(double, double, double)): GFunction(mach)
    {name = nm; fn = f; returnType = g::kScalarValue;}
    virtual void Run() {
        double tmp1 = machine.StackPopScalar();
        double tmp2 = machine.StackPopScalar();
        machine.StackTopScalar() = fn(machine.StackTopScalar(), tmp1, tmp2);
    }
};

//Vector return, vector parameter
class FnV_V: public g::GFunction {
  protected:
    void (*fn)(double *, double, double, double);
  public:
    FnV_V(g::GMachine & mach, const std::string & nm, void (*f)(double *, double, double, double)): GFunction(mach)
    {name = nm; fn = f; returnType = g::kVectorValue;}
    virtual void Run() {
        double * tmp = machine.StackTopVector();
        fn(machine.StackTopVector(), tmp[0], tmp[1], tmp[2]);
    }
};

//Scalar return
class ScalarFunction: public g::GFunction {
  protected:
    void (*fn)(GMachine & mach);
  public:
    ScalarFunction(GMachine & mach, const std::string & nm, void (*f)(GMachine & mach)): GFunction(mach)
    {name = nm; fn = f; returnType = g::kScalarValue;}
    virtual void Run() {fn(machine);}
};

//Vector return
class VectorFunction: public g::GFunction {
  protected:
    void (*fn)(GMachine & mach);
  public:
    VectorFunction(GMachine & mach, const std::string & nm, void (*f)(GMachine & mach)): GFunction(mach)
    {name = nm; fn = f; returnType = g::kVectorValue;}
    virtual void Run() {fn(machine);}
};

typedef unsigned long opcode;
typedef GFunction * GFunctionPtr;

struct GFunctionCode {
    const opcode * codes;
    int codeSize;
    
    //Symbols for parameters
    //(only used here for debugging info)
    GSymbol * parameters;
    int numParameters;
    
    GDataBlock constantData;
    
    size_t stackSize;
    size_t localDataSize;
    size_t parameterDataSize;
    
    GFunctionCode();
    ~GFunctionCode();
};

class GCompiler;
class GFunctionInfo;

class GVMFunction: public GFunction {
  private:
    GFunctionCode * code;
    
  public:
    GVMFunction(GMachine & mach);
    ~GVMFunction();
    
    void Run();
    
    void Dump();
    
    friend class GCompiler;
    friend class GFunctionInfo;
};

}//namespace g

#endif
