/*
 *  gtokenizer.cpp
 *  G
 *
 *  Created by Christopher Huff on Wed Oct 16 2002.
 *  Copyright (c) 2002 Christopher Huff. All rights reserved.
 *
 */

#include "gtokenizer.h"
#include <string>

namespace g {

//This array associates token values with specific strings.
//The tokenizer grabs a string it considers a token and then
//does a linear search on this list for a match.
//Could optimize by sorting and doing a binary search or by
//using a STL container, but this doesn't consume much time.
//May change to a STL container anyway, to allow the tokenizer
//to be used with different keyword sets without modifying this
//file.
Token tokenList[kEndOfCodeToken] = {
    Token(kLCurlyBracketToken, "{"),
    Token(kRCurlyBracketToken, "}"),
    Token(kLParenthesesToken, "("),
    Token(kRParenthesesToken, ")"),
    Token(kLAngleBracketToken, "<"),
    Token(kRAngleBracketToken, ">"),
    Token(kLSquareBracketToken, "["),
    Token(kRSquareBracketToken, "]"),
    Token(kEqualToken, "="),
    Token(kPlusToken, "+"),
    Token(kMinusToken, "-"),
    Token(kStarToken, "*"),
    Token(kSlashToken, "/"),
    Token(kPercentToken, "%"),
    Token(kHatToken, "^"),
    Token(kAmpersandToken, "&"),
    Token(kPipeToken, "|"),
    
    Token(kSemicolonToken, ";"),
    Token(kCommaToken, ","),
    Token(kDotToken, "."),
    
    Token(kScalarToken, "scalar"),
    Token(kVectorToken, "vector"),
    Token(kFunctionToken, "function"),
    
    Token(kWhileToken, "while"),
    Token(kForToken, "for"),
    Token(kDoToken, "do"),
    Token(kIfToken, "if"),
    Token(kElseToken, "else"),
    Token(kReturnToken, "return"),
    
    Token(kIdentifierToken, "identifier"),
    Token(kScalarValueToken, "scalar value"),
    
    Token(kPrintToken, "print"),
    Token(kDumpToken, "dump"),
    Token(kLabelToken, "label")
};


//GTokenizer constructor
//Takes a string containing the source code to be tokenized.
GTokenizer::GTokenizer(const std::string & source)
{
    tokIdx = 0;
    strIdx = 0;
    done = false;
    strSize = source.size();
    sourceString = &source;
    string buffer;
    while(strIdx < strSize)
    {
        buffer.erase();
        EatWhite();
        char curChar = source[strIdx];
        if(isdigit(curChar))
        {
            do
            {
                buffer += curChar;
                strIdx++;
                curChar = source[strIdx];
            }
            while(strIdx < strSize && (isdigit(curChar) || curChar == '.'));
            
            tokens.push_back(Token(kScalarValueToken, buffer));
        }
        else if(isalpha(curChar))
        {
            do
            {
                buffer += curChar;
                strIdx++;
                curChar = source[strIdx];
            }
            while(strIdx < strSize && (isdigit(curChar) || isalpha(curChar) || curChar == '_'));
            
            int j = 0;
            for(; j < kEndOfCodeToken; j++)
            {
                if(tokenList[j].name == buffer)
                {
                    tokens.push_back(Token(tokenList[j].id, buffer));
                    break;
                }
            }
            
            if(j == kEndOfCodeToken)
                tokens.push_back(Token(kIdentifierToken, buffer));
        }
        else
        {
            switch(curChar)
            {
                case '(': tokens.push_back(Token(kLParenthesesToken, "(")); break;
                case ')': tokens.push_back(Token(kRParenthesesToken, ")")); break;
                case '{': tokens.push_back(Token(kLCurlyBracketToken, "{")); break;
                case '}': tokens.push_back(Token(kRCurlyBracketToken, "}")); break;
                case '<': tokens.push_back(Token(kLAngleBracketToken, "<")); break;
                case '>': tokens.push_back(Token(kRAngleBracketToken, ">")); break;
                case '[': tokens.push_back(Token(kLSquareBracketToken, "[")); break;
                case ']': tokens.push_back(Token(kRSquareBracketToken, "]")); break;
                
                case '=': tokens.push_back(Token(kEqualToken, "=")); break;
                case '+': tokens.push_back(Token(kPlusToken, "+")); break;
                case '-': tokens.push_back(Token(kMinusToken, "-")); break;
                case '*': tokens.push_back(Token(kStarToken, "*")); break;
                case '/': tokens.push_back(Token(kSlashToken, "/")); break;
                
                case ';': tokens.push_back(Token(kSemicolonToken, ";")); break;
                case ',': tokens.push_back(Token(kCommaToken, ",")); break;
                case '.': tokens.push_back(Token(kDotToken, ".")); break;
                
                default:
                    cout << "Error in tokenizer: unrecognized character \"" << curChar << "\"\n";
                    exit(-1);
            }
            strIdx++;
        }
    }
    
    tokens.push_back(Token(kEndOfCodeToken, "~~~end of code~~~"));
    tokens.push_back(Token(kEndOfCodeToken, "~~~end of code~~~"));
    
    currentToken = &tokens[0];
    nextToken = &tokens[1];
    
/*    int n = tokens.size();
    for(int j = 0; j < n; j++)
        cout << tokens[j].name << endl;*/
}

GTokenizer::~GTokenizer()
{
}

//Sets the current token to the next available one.
void GTokenizer::GetNextToken()
{
    if(!done)
    {
        tokIdx++;
        currentToken = nextToken;
        nextToken = &tokens[tokIdx + 1];
        if(currentToken->id == kEndOfCodeToken)
            done = true;
    }
}

//Skips over whitespace: spaces, tabs, newlines, returns.
void GTokenizer::EatWhite()
{
    while((*sourceString)[strIdx] == ' '
       || (*sourceString)[strIdx] == '\t'
       || (*sourceString)[strIdx] == '\n'
       || (*sourceString)[strIdx] == '\r')
        strIdx++;
}

//Does a linear search of the token list for a match to a
//token ID value.
string GTokenizer::StringForToken(int tok)
{
    for(int j = 0; j < kEndOfCodeToken; j++)
        if(tokenList[j].id == tok)
            return tokenList[j].name;
    
    return "no token";
}

}//namespace g
